function [genV, genD, genJ] = load_genomic_segments(varargin)

fmt = 'struct';
by_alleles = false;

n_varargs = length (varargin);
expect_arg = false;
for i = 1:n_varargs
    arg = varargin{i};
    if ~expect_arg && ischar (arg)
        switch arg
            case 'alleles',
                by_alleles = true;
            case 'genes',
                by_alleles = false;
            case 'cell',
                fmt = 'cell';
            otherwise, error ('list_assignments:unknownarg', 'Unknown argument %q.', arg);
        end
    else
        if ~expect_arg
            error ('list_assignments:unexparg', 'Unexpected argument at position %d in varargin.', i);
        end
        expect_arg = false;
    end
end


%% load allele sequences

S=fastaread('genes/genomicVs_with_primers.fasta');
n_alleles = length(S);
h_info = regexp({S.Header},'\|','split');

names = cell(length(S),1);
seqs = upper({S.Sequence});
for i=1:length(S)
    names{i}= strtrim( h_info{i}{2});
end

names_split = regexp(names,'*','split');
names_split = reshape([names_split{:}],2,n_alleles);
genes = names_split(1,:);
alleles = names_split(2,:);


V_gene_info = load_data('genes/V_gene_with_primers.info');
n_genes = length(V_gene_info.Gene);
fasta_order = zeros(n_alleles,1);
for i=1:n_alleles
   % fasta_order(i) = V_gene_info.Position(find(strcmp(V_gene_info.Gene, genes{i}),1));
   fasta_order(i) = find(strcmp(V_gene_info.Gene, genes{i}),1);
end
[gene_pos, reorder] = sort(fasta_order);

seqs = seqs(reorder);
names = names(reorder);
genes = genes(reorder);
alleles = alleles(reorder);

if by_alleles
    genV.seq = seqs';
    genV.name = names;
    genV.gene = genes';
    genV.gene_index = gene_pos;
        
    genV.n_alleles = zeros(n_alleles,1);
    genV.allele = zeros(n_alleles,1);
    for i=1:n_alleles
        this_gene_alleles = genV.gene_index==genV.gene_index(i);
        genV.n_alleles(i) = sum(this_gene_alleles);
        genV.allele(i) = sum(this_gene_alleles(1:i-1)) + 1;
    end
    
    genV.Status = V_gene_info.Status(gene_pos);
    genV.Heptamer = V_gene_info.Heptamer(gene_pos);
    genV.Nonamer = V_gene_info.Nonamer(gene_pos);
    genV.Spacer = V_gene_info.Spacer(gene_pos);
    genV.Primer = V_gene_info.Primer(gene_pos);
    
    genV.Heptamer_Index = V_gene_info.Heptamer_Index(gene_pos);
    genV.Nonamer_Index = V_gene_info.Nonamer_Index(gene_pos);
    genV.Position = V_gene_info.Position(gene_pos);
    genV.Orientation = V_gene_info.Orientation(gene_pos);
        
    
else
    
    genV(n_genes).gene = genes{end};
    genV = genV';
    
    for g=1:n_genes
        
        this_gene_alleles = gene_pos==g;
        genV(g).gene = genes{find(this_gene_alleles,1)};
        genV(g).n_alleles = sum(this_gene_alleles);
        genV(g).seq = seqs(this_gene_alleles);
        genV(g).name = names(this_gene_alleles);
        genV(g).alleles = alleles(this_gene_alleles);
        
        genV(g).Status = V_gene_info.Status{g};
        genV(g).Heptamer_Index = V_gene_info.Heptamer_Index(g);
        genV(g).Heptamer = V_gene_info.Heptamer{g};
        genV(g).Nonamer_Index = V_gene_info.Nonamer_Index(g);
        genV(g).Nonamer = V_gene_info.Nonamer{g};
        genV(g).Spacer = V_gene_info.Spacer{g};
        genV(g).Primer = V_gene_info.Primer{g};
        genV(g).Position = V_gene_info.Position(g);
        genV(g).Orientation = V_gene_info.Orientation{g};
        
    end
end



%%



S=fastaread('genes/genomicDs.fasta');

n_alleles = length(S);
seqs = upper({S.Sequence});
names = cell(length(S),1);

for i=1:length(S)
    names{i}= strtrim(strtok(S(i).Header,'|'));
end

names_split = regexp(names,'*','split');
names_split = reshape([names_split{:}],2,n_alleles);
genes = names_split(1,:);
alleles = names_split(2,:);

unique_genes = unique(genes);
n_genes = length(unique_genes);

if by_alleles
    genD.seq = seqs';
    genD.name = names;
    genD.gene = genes';
    genD.gene_index = zeros(n_alleles,1);
    
    for i=1:n_alleles
        genD.gene_index(i) = find(strcmp(unique_genes, genD.gene(i)));
    end
    
    genD.n_alleles = zeros(n_alleles,1);
    genD.allele = zeros(n_alleles,1);
    for i=1:n_alleles
        this_gene_alleles = genD.gene_index==genD.gene_index(i);
        genD.n_alleles(i) = sum(this_gene_alleles);
        genD.allele(i) = sum(this_gene_alleles(1:i-1)) + 1;
    end
    
else
    gene_index = zeros(n_alleles,1);
    genD(n_genes).gene = unique_genes{end};
    genD = genD';
    for i=1:n_alleles
        gene_index(i) = find(strcmp(unique_genes, genes{i}));
        
    end
    
    for g=1:n_genes
        this_gene_alleles = gene_index==g;
        genD(g).gene = unique_genes{g};
        genD(g).n_alleles = sum(this_gene_alleles);
        genD(g).seq = seqs(this_gene_alleles);
        genD(g).name = names(this_gene_alleles);
        genD(g).alleles = alleles(this_gene_alleles);
    end
    
end


%%

S = fastaread('genes/genomicJs_all_curated.fasta');

J_until_primer = load_data('genes/J_gene_with_primers.info');

h_info = regexp({S.Header},'\|','split');
n_alleles = length(S);
seqs = upper({S.Sequence});
names = cell(length(S),1);
status = cell(length(S),1);

for i=1:length(S)
    names{i}= strtrim( h_info{i}{2});
    status{i} = strtrim(h_info{i}(4));
end

names_split = regexp(names,'*','split');
names_split = reshape([names_split{:}],2,n_alleles);
genes = names_split(1,:);
alleles = names_split(2,:);

unique_genes = unique(genes);
n_genes = length(unique_genes);


if by_alleles
    genJ.seq = seqs';
    genJ.name = names;
    genJ.gene = genes';
    genJ.gene_index = zeros(n_alleles,1);
    
    for i=1:n_alleles
        genJ.gene_index(i) = find(strcmp(unique_genes, genJ.gene(i)));
    end
    
    genJ.n_alleles = zeros(n_alleles,1);
    genJ.allele = zeros(n_alleles,1);
    genJ.seq_until_primer = cell(n_alleles,1);
    
    for i=1:n_alleles
        this_gene_alleles = genJ.gene_index==genJ.gene_index(i);
        genJ.n_alleles(i) = sum(this_gene_alleles);
        genJ.allele(i) = sum(this_gene_alleles(1:i-1)) + 1;
        genJ.seq_until_primer{i} = J_until_primer.Sequence_Until_Primer{genJ.gene_index(i)};
    end
    
else
    gene_index = zeros(n_alleles,1);
    genJ(n_genes).gene = unique_genes{end};
    genJ = genJ';
    for i=1:n_alleles
        gene_index(i) = find(strcmp(unique_genes, genes{i}));
        
    end
    
    for g=1:n_genes
        this_gene_alleles = gene_index==g;
        genJ(g).gene = unique_genes{g};
        genJ(g).n_alleles = sum(this_gene_alleles);
        genJ(g).seq = seqs(this_gene_alleles);
        genJ(g).name = names(this_gene_alleles);
        genJ(g).alleles = alleles(this_gene_alleles);
        genJ(g).seq_until_primer = J_until_primer.Sequence_Until_Primer{g};
    end

    
end


%%

if strcmp(fmt,'struct') && by_alleles
    genV = struct_of_arrays_to_array_of_structs(genV);
    genD = struct_of_arrays_to_array_of_structs(genD);
    genJ = struct_of_arrays_to_array_of_structs(genJ);
end



end